@@ -39,6 +39,10 @@
 lonlat_to_cell <- function(lon, lat, grid) {
   g <- extract_grid(grid)
 
+  if (is_h3_grid(g)) {
+    return(cpp_h3_latLngToCell(as.numeric(lon), as.numeric(lat), g@resolution))
+  }
+
   if (g@aperture == "4/3") {
     level <- as.integer(g@resolution / 2)
     cpp_lonlat_to_cell_ap43(
@@ -76,6 +80,11 @@ lonlat_to_cell <- function(lon, lat, grid) {
 cell_to_lonlat <- function(cell_id, grid) {
   g <- extract_grid(grid)
 
+  if (is_h3_grid(g)) {
+    result <- cpp_h3_cellToLatLng(as.character(cell_id))
+    return(data.frame(lon_deg = result$lon, lat_deg = result$lat))
+  }
+
   if (g@aperture == "4/3") {
     level <- as.integer(g@resolution / 2)
     cpp_cell_to_lonlat_ap43(
@@ -100,6 +109,9 @@ cell_to_lonlat <- function(cell_id, grid) {
 #'   uses cells from x.
 #' @param grid A HexGridInfo or HexData object. If HexData and cell_id is NULL,
 #'   polygons are generated for all cells in the data.
+#' @param wrap_dateline Logical. If TRUE (default), calls
+#'   \code{sf::st_wrap_dateline()} to split antimeridian-crossing polygons.
+#'   Set to FALSE for orthographic/globe projections where wrapping creates gaps.
 #'
 #' @return sf object with cell_id and geometry columns
 #'
@@ -122,7 +134,7 @@ cell_to_lonlat <- function(cell_id, grid) {
 #' df <- data.frame(lon = c(0, 10, 20), lat = c(45, 50, 55))
 #' result <- hexify(df, lon = "lon", lat = "lat", area_km2 = 1000)
 #' polys <- cell_to_sf(grid = result)
-cell_to_sf <- function(cell_id = NULL, grid) {
+cell_to_sf <- function(cell_id = NULL, grid, wrap_dateline = TRUE) {
   if (!requireNamespace("sf", quietly = TRUE)) {
     stop("Package 'sf' is required. Install with: install.packages('sf')")
   }
@@ -146,7 +158,34 @@ cell_to_sf <- function(cell_id = NULL, grid) {
     stop("No valid cell_id values")
   }
 
-  # Generate polygons using C++ function
+  # H3 path: use native C backend for boundaries
+  if (is_h3_grid(g)) {
+    boundaries <- cpp_h3_cellToBoundary(as.character(cell_id))
+    polygons <- lapply(boundaries, function(coords) {
+      if (nrow(coords) == 0) return(sf::st_polygon())
+      lons <- coords[, 1]
+      lon_range <- max(lons, na.rm = TRUE) - min(lons, na.rm = TRUE)
+      if (lon_range > 180) {
+        lons[lons < 0] <- lons[lons < 0] + 360
+        coords[, 1] <- lons
+        mean_lon <- mean(lons)
+        if (mean_lon > 180) {
+          coords[, 1] <- coords[, 1] - 360
+        }
+      }
+      sf::st_polygon(list(coords))
+    })
+    sfc <- sf::st_sfc(polygons, crs = g@crs)
+    sfc <- suppressWarnings(sf::st_make_valid(sfc))
+    result_sf <- sf::st_sf(cell_id = as.character(cell_id), geometry = sfc)
+    if (wrap_dateline) {
+      result_sf <- sf::st_wrap_dateline(result_sf,
+        options = c("WRAPDATELINE=YES", "DATELINEOFFSET=180"), quiet = TRUE)
+    }
+    return(result_sf)
+  }
+
+  # ISEA path: generate polygons using C++ function
   # Convert aperture to integer for C++ (mixed aperture "4/3" uses 3)
   aperture_int <- if (g@aperture == "4/3") 3L else as.integer(g@aperture)
 
@@ -156,19 +195,46 @@ cell_to_sf <- function(cell_id = NULL, grid) {
     aperture_int
   )
 
+
+  # Handle antimeridian-crossing polygons: normalize coordinates so each
+  # polygon is contiguous. When wrap_dateline = TRUE, st_wrap_dateline
+  # then splits at ±180° for correct flat-map rendering. For globe/
+  # orthographic projections, pass wrap_dateline = FALSE to keep cells intact.
+
   polygons <- lapply(corners_list, function(coords) {
+    lons <- coords[, 1]
+    lon_range <- max(lons, na.rm = TRUE) - min(lons, na.rm = TRUE)
+
+    if (lon_range > 180) {
+      # Polygon crosses antimeridian - normalize to be contiguous
+      # Shift negative lons to 0-360 range
+      lons[lons < 0] <- lons[lons < 0] + 360
+      coords[, 1] <- lons
+
+      # Now shift back to standard range, but keeping contiguity
+      # If centroid is > 180, shift everything by -360
+      mean_lon <- mean(lons)
+      if (mean_lon > 180) {
+        coords[, 1] <- coords[, 1] - 360
+      }
+    }
+
     sf::st_polygon(list(coords))
   })
 
   sfc <- sf::st_sfc(polygons, crs = g@crs)
 
-  # Handle antimeridian-crossing polygons by wrapping at the dateline
+  # Fix any invalid geometries (self-intersecting polygons, etc.)
+  # suppressWarnings: antimeridian normalization may temporarily produce
+  # out-of-range longitudes that st_wrap_dateline corrects below
+  sfc <- suppressWarnings(sf::st_make_valid(sfc))
 
-  # This splits polygons that cross ±180° longitude into valid MULTIPOLYGONs
-  # DATELINEOFFSET=180 ensures proper splitting at the antimeridian
-  sfc <- sf::st_wrap_dateline(sfc, options = c("WRAPDATELINE=YES", "DATELINEOFFSET=180"))
-
-  sf::st_sf(cell_id = cell_id, geometry = sfc)
+  result_sf <- sf::st_sf(cell_id = cell_id, geometry = sfc)
+  if (wrap_dateline) {
+    result_sf <- sf::st_wrap_dateline(result_sf,
+      options = c("WRAPDATELINE=YES", "DATELINEOFFSET=180"), quiet = TRUE)
+  }
+  result_sf
 }
 
 # =============================================================================
@@ -178,6 +244,8 @@ cell_to_sf <- function(cell_id = NULL, grid) {
 #' Generate a rectangular grid of hexagons
 #'
 #' Creates hexagon polygons covering a rectangular geographic region.
+#' For H3 grids, all cells that overlap the bounding box are included
+#' (not just cells whose center falls inside), ensuring full spatial coverage.
 #'
 #' @param bbox Bounding box as c(xmin, ymin, xmax, ymax), or an sf/sfc object
 #' @param grid A HexGridInfo object specifying the grid parameters
@@ -203,6 +271,22 @@ grid_rect <- function(bbox, grid) {
     bbox <- as.numeric(sf::st_bbox(bbox))
   }
 
+  # H3 path: fill bbox with cells using native C backend
+  if (is_h3_grid(g)) {
+    bbox_coords <- matrix(c(
+      bbox[1], bbox[2],
+      bbox[3], bbox[2],
+      bbox[3], bbox[4],
+      bbox[1], bbox[4],
+      bbox[1], bbox[2]
+    ), ncol = 2, byrow = TRUE)
+    cell_ids <- cpp_h3_polygonToCells(bbox_coords, g@resolution)
+    if (length(cell_ids) == 0) {
+      stop("No H3 cells found in the specified bounding box at resolution ", g@resolution)
+    }
+    return(cell_to_sf(cell_ids, g))
+  }
+
   minlon <- bbox[1]
   minlat <- bbox[2]
   maxlon <- bbox[3]
@@ -228,6 +312,9 @@ grid_rect <- function(bbox, grid) {
 #' Creates hexagon polygons covering the entire Earth.
 #'
 #' @param grid A HexGridInfo object specifying the grid parameters
+#' @param wrap_dateline Logical. If TRUE (default), antimeridian-crossing
+#'   polygons are split at +/-180 degrees. Set to FALSE for orthographic/globe
+#'   projections where wrapping creates gaps.
 #'
 #' @return sf object with hexagon polygons
 #'
@@ -244,14 +331,39 @@ grid_rect <- function(bbox, grid) {
 #' grid <- hex_grid(area_km2 = 100000)
 #' global <- grid_global(grid)
 #' plot(global)
-grid_global <- function(grid) {
+grid_global <- function(grid, wrap_dateline = TRUE) {
   if (!requireNamespace("sf", quietly = TRUE)) {
     stop("Package 'sf' is required")
   }
 
   g <- extract_grid(grid)
 
-  # Estimate cell count for warning
+  # H3 path: fill globe using native C backend
+  if (is_h3_grid(g)) {
+    h3_n_cells <- 2 + 120 * 7^g@resolution
+    if (h3_n_cells > 2e6) {
+      warning(sprintf(
+        "H3 global grid at res %d has ~%.0f cells. This may take a while.",
+        g@resolution, h3_n_cells
+      ))
+    }
+    # Split globe into quadrants for polygonToCells
+    quads <- list(
+      matrix(c(-180, 0, 0, 0, 0, 90, -180, 90, -180, 0), ncol = 2, byrow = TRUE),
+      matrix(c(0, 0, 180, 0, 180, 90, 0, 90, 0, 0), ncol = 2, byrow = TRUE),
+      matrix(c(-180, -90, 0, -90, 0, 0, -180, 0, -180, -90), ncol = 2, byrow = TRUE),
+      matrix(c(0, -90, 180, -90, 180, 0, 0, 0, 0, -90), ncol = 2, byrow = TRUE)
+    )
+    all_cells <- character(0)
+    for (q in quads) {
+      quad_cells <- cpp_h3_polygonToCells(q, g@resolution)
+      all_cells <- c(all_cells, quad_cells)
+    }
+    cell_ids <- unique(all_cells)
+    return(cell_to_sf(cell_ids, g, wrap_dateline = wrap_dateline))
+  }
+
+  # Estimate cell count for warning (ISEA)
   if (g@aperture == "4/3") {
     level <- as.integer(g@resolution / 2)
     n_cells <- 10 * (4^level) * (3^(g@resolution - level)) + 2
@@ -276,8 +388,12 @@ grid_global <- function(grid) {
 
   # Add polar cap sampling (±85 to ±90 degrees)
   # The regular grid misses polar cells because lat stops at ±85
-  polar_lons <- seq(-180, 180, by = 30)  # Coarser sampling near poles is fine
-  polar_lats <- c(seq(85.5, 89.9, by = 1), seq(-89.9, -85.5, by = 1))
+  # Near poles, longitude spacing must be DENSER not coarser - at 89°N,
+  # the entire circumference is only ~6.3° of longitude-equivalent distance.
+  # Use the same spacing_deg (or denser) to ensure we catch all cells.
+  polar_lon_spacing <- min(spacing_deg, 15)  # At most 15°, or cell-based spacing
+  polar_lons <- seq(-180, 180, by = polar_lon_spacing)
+  polar_lats <- c(seq(85.5, 89.99, by = 0.5), seq(-89.99, -85.5, by = 0.5))
   polar_pts <- expand.grid(lon = polar_lons, lat = polar_lats)
 
   # Combine main grid with polar samples
@@ -286,7 +402,7 @@ grid_global <- function(grid) {
   cell_ids <- lonlat_to_cell(grid_pts$lon, grid_pts$lat, g)
   unique_cells <- unique(cell_ids)
 
-  cell_to_sf(unique_cells, g)
+  cell_to_sf(unique_cells, g, wrap_dateline = wrap_dateline)
 }
 
 #' Clip hexagon grid to polygon boundary
@@ -304,8 +420,10 @@ grid_global <- function(grid) {
 #' @return sf object with hexagon polygons clipped to the boundary
 #'
 #' @details
-#' The function first generates a rectangular grid covering the bounding box
-#' of the input polygon, then clips or filters cells to the boundary.
+#' The function first generates cells covering the boundary polygon, then
+#' clips or filters them. For H3 grids, all cells that overlap the boundary
+#' are included (not just cells whose center falls inside), ensuring full
+#' spatial coverage with no gaps along the boundary edge.
 #'
 #' When \code{crop = TRUE}, hexagons are geometrically intersected with the
 #' boundary, which may produce partial hexagons at the edges. When
@@ -348,6 +466,48 @@ grid_clip <- function(boundary, grid, crop = TRUE) {
 
   g <- extract_grid(grid)
 
+  # H3 path: fill boundary polygon using native C backend
+  if (is_h3_grid(g)) {
+    # Disable S2 for spatial operations
+    s2_state <- sf::sf_use_s2()
+    sf::sf_use_s2(FALSE)
+    on.exit(sf::sf_use_s2(s2_state), add = TRUE)
+
+    boundary_geom <- sf::st_geometry(boundary)
+    if (length(boundary_geom) > 1) {
+      boundary_geom <- sf::st_union(boundary_geom)
+    }
+    boundary_geom <- sf::st_make_valid(boundary_geom)
+
+    # Extract polygon rings for cpp_h3_polygonToCells
+    polys <- sf::st_cast(boundary_geom, "POLYGON")
+    all_cells <- character(0)
+    for (p in polys) {
+      rings <- unclass(p)
+      outer_ring <- rings[[1]]
+      hole_rings <- if (length(rings) > 1) rings[-1] else NULL
+      pcells <- cpp_h3_polygonToCells(outer_ring, g@resolution, holes = hole_rings)
+      all_cells <- c(all_cells, pcells)
+    }
+    cell_ids <- unique(all_cells)
+    hex_sf <- cell_to_sf(cell_ids, g)
+
+    if (crop) {
+      hex_sf <- sf::st_make_valid(hex_sf)
+      result <- tryCatch({
+        suppressWarnings(sf::st_intersection(hex_sf, boundary_geom))
+      }, error = function(e) {
+        boundary_buf <- sf::st_buffer(boundary_geom, 0)
+        hex_buf <- sf::st_buffer(hex_sf, 0)
+        suppressWarnings(sf::st_intersection(hex_buf, boundary_buf))
+      })
+      geom_types <- sf::st_geometry_type(result)
+      result <- result[geom_types %in% c("POLYGON", "MULTIPOLYGON"), ]
+      return(result)
+    }
+    return(hex_sf)
+  }
+
   # Get bounding box of boundary
   bbox <- sf::st_bbox(boundary)
 
@@ -393,6 +553,82 @@ grid_clip <- function(boundary, grid, crop = TRUE) {
   result
 }
 
+# =============================================================================
+# CELL AREA COMPUTATION
+# =============================================================================
+
+#' Compute per-cell area in km²
+#'
+#' Returns the area of each cell in square kilometers. For ISEA grids, all
+#' cells have the same area (equal-area property). For H3 grids, each cell
+#' has a different geodesic area depending on its location.
+#'
+#' @param cell_id Cell IDs to compute area for. For ISEA grids, these are
+#'   numeric; for H3 grids, character strings. When \code{grid} is a HexData
+#'   object and \code{cell_id} is \code{NULL}, all cell IDs from the data are
+#'   used.
+#' @param grid A HexGridInfo or HexData object.
+#'
+#' @return Named numeric vector of areas in km², one per \code{cell_id}.
+#'
+#' @details
+#' For ISEA grids the area is constant across all cells and is read directly
+#' from the grid specification.
+#'
+#' For H3 grids the area varies by latitude. This function computes geodesic
+#' area via \code{sf::st_area()} on H3 cell polygons, with results cached in a
+#' session-scoped environment so repeated calls for the same cells are fast.
+#'
+#' @seealso \code{\link{hex_grid}} for grid specifications,
+#'   \code{\link{h3_crosswalk}} for ISEA/H3 interoperability
+#'
+#' @export
+#' @examples
+#' # ISEA: constant area
+#' grid <- hex_grid(area_km2 = 1000)
+#' cells <- lonlat_to_cell(c(0, 10, 20), c(45, 50, 55), grid)
+#' cell_area(cells, grid)
+#'
+#' # H3: area varies by location
+#' \donttest{
+#' h3 <- hex_grid(resolution = 5, type = "h3")
+#' h3_cells <- lonlat_to_cell(c(0, 0), c(0, 80), h3)
+#' cell_area(h3_cells, h3)  # equator vs polar — different areas
+#' }
+cell_area <- function(cell_id = NULL, grid) {
+
+  # Handle HexData input
+  if (is_hex_data(grid)) {
+    if (is.null(cell_id)) {
+      cell_id <- grid@cell_id
+    }
+    g <- grid@grid
+  } else {
+    g <- extract_grid(grid)
+    if (is.null(cell_id)) {
+      stop("cell_id required when grid is not HexData")
+    }
+  }
+
+  # ISEA: constant equal-area
+  if (!is_h3_grid(g)) {
+    areas <- rep(g@area_km2, length(cell_id))
+    if (is.numeric(cell_id)) {
+      names(areas) <- as.character(as.integer(cell_id))
+    } else {
+      names(areas) <- as.character(cell_id)
+    }
+    return(areas)
+  }
+
+  # H3: per-cell area via native C backend
+  cell_id <- as.character(cell_id)
+  areas <- cpp_h3_cellAreaKm2(cell_id)
+  names(areas) <- cell_id
+  areas
+}
+
+
 # =============================================================================
 # HIERARCHICAL INDEX HELPERS
 # =============================================================================
@@ -412,6 +648,11 @@ grid_clip <- function(boundary, grid, crop = TRUE) {
 cell_to_index <- function(cell_id, grid) {
   g <- extract_grid(grid)
 
+  # H3 cell IDs are already hierarchical index strings
+  if (is_h3_grid(g)) {
+    return(as.character(cell_id))
+  }
+
   # Determine index type based on aperture
   index_type <- if (g@aperture == "3") "z3"
                 else if (g@aperture == "7") "z7"
@@ -455,6 +696,15 @@ get_parent <- function(cell_id, grid, levels = 1L) {
     stop("Cannot get parent: already at minimum resolution")
   }
 
+  # H3 path
+  if (is_h3_grid(g)) {
+    parent_res <- g@resolution - as.integer(levels)
+    if (parent_res < H3_MIN_RESOLUTION) {
+      stop("Cannot get parent: would go below H3 minimum resolution")
+    }
+    return(cpp_h3_cellToParent(as.character(cell_id), parent_res))
+  }
+
   index_type <- if (g@aperture == "3") "z3"
                 else if (g@aperture == "7") "z7"
                 else "zorder"
@@ -499,6 +749,15 @@ get_parent <- function(cell_id, grid, levels = 1L) {
 get_children <- function(cell_id, grid, levels = 1L) {
   g <- extract_grid(grid)
 
+  # H3 path
+  if (is_h3_grid(g)) {
+    child_res <- g@resolution + as.integer(levels)
+    if (child_res > H3_MAX_RESOLUTION) {
+      stop("Cannot get children: would exceed H3 maximum resolution (15)")
+    }
+    return(cpp_h3_cellToChildren(as.character(cell_id), child_res))
+  }
+
   if (g@resolution + levels > MAX_RESOLUTION) {
     stop("Cannot get children: would exceed maximum resolution")
   }
